# -*- coding: utf-8 -*-

from __future__ import absolute_import

from benedict.serializers.abstract import AbstractSerializer

try:
    # python 3
    from urllib.parse import urlencode
    from urllib.parse import parse_qs
except ImportError:
    # python 2
    from urllib import urlencode
    from urlparse import parse_qs

import re


class QueryStringSerializer(AbstractSerializer):

    def __init__(self):
        super(QueryStringSerializer, self).__init__()

    def decode(self, s, **kwargs):
        flat = kwargs.pop('flat', True)
        qs_re = r'^(([\w\-\%\+]+\=[\w\-\%\+]*)+([\&]{1})?)+'
        qs_pattern = re.compile(qs_re)
        if qs_pattern.match(s):
            data = parse_qs(s)
            if flat:
                data = {key: value[0] for key, value in data.items()}
            return data
        raise ValueError('Invalid query string: {}'.format(s))

    def encode(self, d, **kwargs):
        data = urlencode(d, **kwargs)
        return data



# -*- coding: utf-8 -*-

from benedict.serializers import JSONSerializer


def dump(obj, **kwargs):
    serializer = JSONSerializer()
    options = {
        'indent': 4,
        'sort_keys': True,
    }
    options.update(**kwargs)
    try:
        output = serializer.encode(obj, **options)
        return output
    except TypeError as error:
        sort_keys = options.pop('sort_keys', False)
        if sort_keys:
            output = serializer.encode(obj, **options)
            return output
        raise error



# -*- coding: utf-8 -*-

from benedict.core import clean as _clean
from benedict.core import clone as _clone
from benedict.core import dump as _dump
from benedict.core import filter as _filter
from benedict.core import find as _find
from benedict.core import flatten as _flatten
from benedict.core import groupby as _groupby
from benedict.core import invert as _invert
from benedict.core import items_sorted_by_keys as _items_sorted_by_keys
from benedict.core import items_sorted_by_values as _items_sorted_by_values
from benedict.core import keypaths as _keypaths
from benedict.core import match as _match
from benedict.core import merge as _merge
from benedict.core import move as _move
from benedict.core import nest as _nest
from benedict.core import remove as _remove
from benedict.core import rename as _rename
from benedict.core import search as _search
from benedict.core import standardize as _standardize
from benedict.core import subset as _subset
from benedict.core import swap as _swap
from benedict.core import traverse as _traverse
from benedict.core import unflatten as _unflatten
from benedict.core import unique as _unique
from benedict.dicts.io import IODict
from benedict.dicts.keylist import KeylistDict
from benedict.dicts.keypath import KeypathDict
from benedict.dicts.parse import ParseDict


class benedict(KeypathDict, IODict, ParseDict):

    def __init__(self, *args, **kwargs):
        """
        Constructs a new instance.
        """
        if len(args) == 1 and isinstance(args[0], benedict):
            obj = args[0]
            kwargs.setdefault('keypath_separator', obj.keypath_separator)
            super(benedict, self).__init__(obj.dict(), **kwargs)
            return
        super(benedict, self).__init__(*args, **kwargs)

    def __getitem__(self, key):
        return self._cast(
            super(benedict, self).__getitem__(key))

    def _cast(self, value):
        """
        Cast a dict instance to a benedict instance
        keeping the pointer to the original dict.
        """
        if isinstance(value, dict) and not isinstance(value, benedict):
            return benedict(value, keypath_separator=self._keypath_separator)
        return value

    def clean(self, strings=True, collections=True):
        """
        Clean the current dict instance removing all empty values: None, '', {}, [], ().
        If strings or collections (dict, list, set, tuple) flags are False,
        related empty values will not be deleted.
        """
        _clean(self, strings=strings, collections=collections)

    def clone(self):
        """
        Creates and return a clone of the current dict instance (deep copy).
        """
        return self._cast(
            _clone(self))

    def copy(self):
        """
        Creates and return a copy of the current instance (shallow copy).
        """
        return self._cast(
            super(benedict, self).copy())

    def deepcopy(self):
        """
        Alias of 'clone' method.
        """
        return self.clone()

    def deepupdate(self, other, *args):
        """
        Alias of 'merge' method.
        """
        self.merge(other, *args)

    def dump(self, data=None):
        """
        Return a readable string representation of any dict/list.
        This method can be used both as static method or instance method.
        """
        return _dump(data or self)

    def filter(self, predicate):
        """
        Return a new filtered dict using the given predicate function.
        Predicate function receives key, value arguments and should return a bool value.
        """
        return _filter(self, predicate)

    def find(self, keys, default=None):
        """
        Return the first match searching for the given keys.
        If no result found, default value is returned.
        """
        return _find(self, keys, default)

    def flatten(self, separator='_'):
        """
        Return a new flattened dict using the given separator
        to join nested dict keys to flatten keypaths.
        """
        return _flatten(self, separator)

    def get(self, key, default=None):
        return self._cast(
            super(benedict, self).get(key, default))

    def get_dict(self, key, default=None):
        return self._cast(
            super(benedict, self).get_dict(key, default))

    def get_list_item(self, key, index=0, default=None, separator=','):
        return self._cast(
            super(benedict, self).get_list_item(
                key, index, default, separator))

    def groupby(self, key, by_key):
        """
        Group a list of dicts at key by the value of the given by_key and return a new dict.
        """
        return self._cast(
            _groupby(self[key], by_key))

    def invert(self, flat=False):
        """
        Return a new inverted dict, where values become keys and keys become values.
        Since multiple keys could have the same value, each value will be a list of keys.
        If flat is True each value will be a single value (use this only if values are unique).
        """
        return _invert(self, flat)

    def items_sorted_by_keys(self, reverse=False):
        """
        Return items (key/value list) sorted by keys.
        If reverse is True, the list will be reversed.
        """
        return _items_sorted_by_keys(self, reverse=reverse)

    def items_sorted_by_values(self, reverse=False):
        """
        Return items (key/value list) sorted by values.
        If reverse is True, the list will be reversed.
        """
        return _items_sorted_by_values(self, reverse=reverse)

    def keypaths(self, indexes=False):
        """
        Return a list of all keypaths in the dict.
        If indexes is True, the output will include list values indexes.
        """
        return _keypaths(
            self, separator=self._keypath_separator, indexes=indexes)

    def match(self, pattern, indexes=True):
        """
        Return a list of all values whose keypath matches the given pattern (a regex or string).
        If pattern is string, wildcard can be used (eg. [*] can be used to match all list indexes).
        If indexes is True, the pattern will be matched also against list values.
        """
        return _match(
            self, pattern, separator=self._keypath_separator, indexes=indexes)

    def merge(self, other, *args, **kwargs):
        """
        Merge one or more dict objects into current instance (deepupdate).
        Sub-dictionaries will be merged toghether.
        If overwrite is False, existing values will not be overwritten.
        """
        _merge(self, other, *args, **kwargs)

    def move(self, key_src, key_dest):
        """
        Move a dict instance value item from 'key_src' to 'key_dst'.
        If key_dst exists, its value will be overwritten.
        """
        _move(self, key_src, key_dest)

    def nest(self, key,
             id_key='id', parent_id_key='parent_id', children_key='children'):
        """
        Nest a list of dicts at the given key and return a new nested list
        using the specified keys to establish the correct items hierarchy.
        """
        return _nest(self[key], id_key, parent_id_key, children_key)

    def pop(self, key, *args):
        return self._cast(
            super(benedict, self).pop(key, *args))

    def remove(self, keys, *args):
        """
        Remove multiple keys from the current dict instance.
        It is possible to pass a single key or more keys (as list or *args).
        """
        _remove(self, keys, *args)

    def setdefault(self, key, default=None):
        return self._cast(
            super(benedict, self).setdefault(key, default))

    def rename(self, key, key_new):
        """
        Rename a dict item key from 'key' to 'key_new'.
        If key_new exists, a KeyError will be raised.
        """
        _rename(self, key, key_new)

    def search(self, query,
               in_keys=True, in_values=True,
               exact=False, case_sensitive=False):
        """
        Search and return a list of items (dict, key, value, ) matching the given query.
        """
        return _search(self, query, in_keys, in_values, exact, case_sensitive)

    def standardize(self):
        """
        Standardize all dict keys (e.g. 'Location Latitude' -> 'location_latitude').
        """
        _standardize(self)

    def subset(self, keys, *args):
        """
        Return a new dict subset for the given keys.
        It is possible to pass a single key or multiple keys (as list or *args).
        """
        return _subset(self, keys, *args)

    def swap(self, key1, key2):
        """
        Swap items values at the given keys.
        """
        _swap(self, key1, key2)

    def traverse(self, callback):
        """
        Traverse the current dict instance (including nested dicts),
        and pass each item (dict, key, value) to the callback function.
        """
        _traverse(self, callback)

    def unflatten(self, separator='_'):
        """
        Return a new unflattened dict using the given separator
        to split dict keys to nested keypaths.
        """
        return _unflatten(self, separator)

    def unique(self):
        """
        Remove duplicated values from the current dict instance.
        """
        _unique(self)



# -*- coding: utf-8 -*-

from benedict.utils import type_util


def groupby(items, key):
    if not type_util.is_list(items):
        raise ValueError('items should be a list of dicts.')
    items_grouped = {}
    for item in items:
        if not type_util.is_dict(item):
            raise ValueError('item should be a dict.')
        group = item.get(key)
        if group not in items_grouped:
            items_grouped[group] = []
        items_grouped[group].append(item.copy())
    return items_grouped



# -*- coding: utf-8 -*-

from benedict.core import clone
from benedict.utils import type_util


def _flatten_key(base_key, key, separator):
    if base_key and separator:
        return '{}{}{}'.format(base_key, separator, key)
    return key


def _flatten_item(d, base_dict, base_key, separator):
    new_dict = base_dict
    keys = list(d.keys())
    for key in keys:
        new_key = _flatten_key(base_key, key, separator)
        value = d.get(key, None)
        if type_util.is_dict(value):
            new_value = _flatten_item(value,
                                      base_dict=new_dict,
                                      base_key=new_key,
                                      separator=separator)
            new_dict.update(new_value)
            continue
        if new_key in new_dict:
            raise KeyError(
                'Invalid key: "{}", key already in flatten dict.'.format(
                    new_key))
        new_dict[new_key] = value
    return new_dict


def flatten(d, separator='_'):
    new_dict = clone(d, empty=True)
    return _flatten_item(d,
                         base_dict=new_dict,
                         base_key='',
                         separator=separator)



# -*- coding: utf-8 -*-


class BaseDict(dict):

    _dict = {}

    def __init__(self, *args, **kwargs):
        if len(args) == 1 and isinstance(args[0], dict):
            self._dict = args[0]
        else:
            self._dict = dict(*args, **kwargs)
        super(BaseDict, self).__init__(*args, **kwargs)

    def __bool__(self):
        return bool(self._dict)

    def __contains__(self, key):
        return key in self._dict

    def __delitem__(self, key):
        del self._dict[key]
        super(BaseDict, self).__delitem__(key)

    def __eq__(self, other):
        return self._dict == other

    def __getitem__(self, key):
        return self._dict[key]

    def __iter__(self):
        return iter(self._dict)

    def __len__(self):
        return len(self._dict)

    def __nonzero__(self):
        return bool(self._dict)

    def __repr__(self):
        return repr(self._dict)

    def __setitem__(self, key, value):
        self._dict[key] = value
        super(BaseDict, self).__setitem__(key, value)

    def __str__(self):
        return str(self._dict)

    def __unicode__(self):
        return unicode(self._dict)

    def clear(self):
        self._dict.clear()
        super(BaseDict, self).clear()

    def copy(self):
        return self._dict.copy()

    def dict(self):
        return self._dict

    def get(self, key, default=None):
        return self._dict.get(key, default)

    def items(self):
        return self._dict.items()

    def keys(self):
        return self._dict.keys()

    def pop(self, key, *args):
        value = self._dict.pop(key, *args)
        super(BaseDict, self).pop(key, None)
        return value

    def setdefault(self, key, default=None):
        value = self._dict.setdefault(key, default)
        super(BaseDict, self).setdefault(key, default)
        return value

    def update(self, other):
        self._dict.update(other)
        super(BaseDict, self).update(other)

    def values(self):
        return self._dict.values()



# -*- coding: utf-8 -*-


class AbstractSerializer(object):

    def __init__(self):
        super(AbstractSerializer, self).__init__()

    def decode(self, s, **kwargs):
        raise NotImplementedError()

    def encode(self, d, **kwargs):
        raise NotImplementedError()



# -*- coding: utf-8 -*-

from __future__ import absolute_import

from benedict.serializers.abstract import AbstractSerializer

import base64
import pickle


class PickleSerializer(AbstractSerializer):

    def __init__(self):
        super(PickleSerializer, self).__init__()

    def decode(self, s, **kwargs):
        encoding = kwargs.pop('encoding', 'utf-8')
        return pickle.loads(
            base64.b64decode(s.encode(encoding)), **kwargs)

    def encode(self, d, **kwargs):
        encoding = kwargs.pop('encoding', 'utf-8')
        kwargs.setdefault('protocol', 2)
        return base64.b64encode(
            pickle.dumps(d, **kwargs)).decode(encoding)



# -*- coding: utf-8 -*-

from benedict.utils import type_util


def _traverse_collection(d, callback):
    if type_util.is_dict(d):
        _traverse_dict(d, callback)
    elif type_util.is_list_or_tuple(d):
        _traverse_list(d, callback)


def _traverse_dict(d, callback):
    keys = list(d.keys())
    for key in keys:
        value = d.get(key, None)
        callback(d, key, value)
        _traverse_collection(value, callback)


def _traverse_list(l, callback):
    items = list(enumerate(l))
    for index, value in items:
        callback(l, index, value)
        _traverse_collection(value, callback)


def traverse(d, callback):
    if not callable(callback):
        raise ValueError('callback argument must be a callable.')
    _traverse_collection(d, callback)



# -*- coding: utf-8 -*-


def move(d, key_src, key_dest, overwrite=True):
    if key_dest == key_src:
        return
    if key_dest in d and not overwrite:
        raise KeyError(
            'Invalid key: "{}", key already in target dict '
            'and `overwrite` is disabled.'.format(key_dest))
    d[key_dest] = d.pop(key_src)



# -*- coding: utf-8 -*-

from benedict.core import traverse
from benedict.utils import type_util

import re


KEY_INDEX_RE = r'(?:\[[\'\"]*(\-?[\d]+)[\'\"]*\]){1}$'


def check_keys(d, separator):
    """
    Check if dict keys contain keypath separator.
    """
    if not type_util.is_dict(d) or not separator:
        return

    def check_key(parent, key, value):
        if key and type_util.is_string(key) and separator in key:
            raise ValueError(
                'keys should not contain keypath separator '
                '\'{}\', found: \'{}\'.'.format(separator, key))
    traverse(d, check_key)


def parse_keys(keypath, separator):
    """
    Parse keys from keylist or keypath using the given separator.
    """
    if type_util.is_list_or_tuple(keypath):
        keys = []
        for key in keypath:
            keys += parse_keys(key, separator)
        return keys
    return _split_keys_and_indexes(keypath, separator)


def _split_key_indexes(key):
    """
    Splits key indexes:
    eg. 'item[0][1]' -> ['item', 0, 1].
    """
    if '[' in key and key.endswith(']'):
        keys = []
        while True:
            matches = re.findall(KEY_INDEX_RE, key)
            if matches:
                key = re.sub(KEY_INDEX_RE, '', key)
                index = int(matches[0])
                keys.insert(0, index)
                # keys.insert(0, { keylist_util.INDEX_KEY:index })
                continue
            keys.insert(0, key)
            break
        return keys
    return [key]


def _split_keys(keypath, separator):
    """
    Splits keys using the given separator:
    eg. 'item.subitem[1]' -> ['item', 'subitem[1]'].
    """
    if separator:
        return keypath.split(separator)
    return [keypath]


def _split_keys_and_indexes(keypath, separator):
    """
    Splits keys and indexes using the given separator:
    eg. 'item[0].subitem[1]' -> ['item', 0, 'subitem', 1].
    """
    if type_util.is_string(keypath):
        keys1 = _split_keys(keypath, separator)
        keys2 = []
        for key in keys1:
            keys2 += _split_key_indexes(key)
        return keys2
    return [keypath]



# -*- coding: utf-8 -*-

from benedict.core.traverse import traverse
from benedict.utils import type_util


def _get_term(value, case_sensitive):
    v_is_str = type_util.is_string(value)
    v = value.lower() if (v_is_str and not case_sensitive) else value
    return (v, v_is_str, )


def _get_match(query, value, exact, case_sensitive):
    q, q_is_str = _get_term(query, case_sensitive)
    v, v_is_str = _get_term(value, case_sensitive)
    # TODO: add regex support
    if q_is_str and v_is_str and not exact:
        return q in v
    return q == v


def search(d, query,
           in_keys=True, in_values=True, exact=False, case_sensitive=True):
    items = []

    def _search_item(item_dict, item_key, item_value):
        match_key = in_keys and _get_match(
            query, item_key, exact, case_sensitive)
        match_val = in_values and _get_match(
            query, item_value, exact, case_sensitive)
        if any([match_key, match_val]):
            items.append((item_dict, item_key, item_value, ))
    traverse(d, _search_item)
    return items



# -*- coding: utf-8 -*-

from benedict.serializers import JSONSerializer
from benedict.utils import type_util

from datetime import datetime
from dateutil import parser as date_parser
from decimal import Decimal, DecimalException
from MailChecker import MailChecker
from phonenumbers import phonenumberutil, PhoneNumberFormat
from six import text_type
from slugify import slugify

import ftfy
import phonenumbers
import re


def _parse_with(val, type_checker, parser, **kwargs):
    if val is None:
        return None
    if callable(type_checker) and type_checker(val):
        return val
    s = text_type(val)
    if not len(s):
        return None
    return parser(s, **kwargs)


def _parse_bool(val):
    val = val.lower()
    if val in ['1', 'true', 'yes', 'ok', 'on']:
        return True
    elif val in ['0', 'false', 'no', 'ko', 'off']:
        return False
    return None


def parse_bool(val):
    return _parse_with(val, type_util.is_bool, _parse_bool)


def _parse_datetime_with_format(val, format):
    try:
        return datetime.strptime(val, format)
    except Exception:
        return None


def _parse_datetime_without_format(val):
    try:
        return date_parser.parse(val)
    except Exception:
        return _parse_datetime_from_timestamp(val)


def _parse_datetime_from_timestamp(val):
    try:
        return datetime.fromtimestamp(float(val))
    except Exception:
        return None


def parse_datetime(val, format=None):
    if type_util.is_datetime(val):
        return val
    s = text_type(val)
    if format:
        return _parse_datetime_with_format(s, format)
    else:
        return _parse_datetime_without_format(s)


def _parse_decimal(val):
    try:
        return Decimal(val)
    except (ValueError, DecimalException):
        return None


def parse_decimal(val):
    return _parse_with(val, type_util.is_decimal, _parse_decimal)


def _parse_dict(val):
    serializer = JSONSerializer()
    try:
        d = serializer.decode(val)
        if type_util.is_dict(d):
            return d
        return None
    except Exception:
        return None


def parse_dict(val):
    return _parse_with(val, type_util.is_dict, _parse_dict)


def _parse_float(val):
    try:
        return float(val)
    except ValueError:
        return None


def parse_float(val):
    return _parse_with(val, type_util.is_float, _parse_float)


def _parse_email(val, check_blacklist=True):
    val = val.lower()
    if check_blacklist:
        if not MailChecker.is_valid(val):
            return None
    else:
        if not MailChecker.is_valid_email_format(val):
            return None
    return val


def parse_email(val, check_blacklist=True):
    return _parse_with(
        val, None, _parse_email, check_blacklist=check_blacklist)


def _parse_int(val):
    try:
        return int(val)
    except ValueError:
        return None


def parse_int(val):
    return _parse_with(val, type_util.is_integer, _parse_int)


def _parse_list(val, separator=None):
    serializer = JSONSerializer()
    try:
        l = serializer.decode(val)
        if type_util.is_list(l):
            return l
    except Exception:
        if separator:
            l = list(val.split(separator))
            return l
    return None


def parse_list(val, separator=None):
    val = _parse_with(val, type_util.is_list_or_tuple,
                      _parse_list, separator=separator)
    return list(val) if type_util.is_list_or_tuple(val) else val


def _parse_phonenumber(val, country_code=None):
    try:
        phone_obj = phonenumbers.parse(val, country_code)
        if phonenumbers.is_valid_number(phone_obj):
            return {
                'e164': phonenumbers.format_number(
                    phone_obj, PhoneNumberFormat.E164),
                'international': phonenumbers.format_number(
                    phone_obj, PhoneNumberFormat.INTERNATIONAL),
                'national': phonenumbers.format_number(
                    phone_obj, PhoneNumberFormat.NATIONAL),
            }
        return None
    except phonenumberutil.NumberParseException:
        return None


def parse_phonenumber(val, country_code=None):
    s = parse_str(val)
    if not s:
        return None
    phone_raw = re.sub(r'[^0-9\+]', ' ', s)
    phone_raw = phone_raw.strip()
    if phone_raw.startswith('00'):
        phone_raw = '+{}'.format(phone_raw[2:])
    if country_code and len(country_code) >= 2:
        country_code = country_code[0:2].upper()
    return _parse_with(
        phone_raw, None, _parse_phonenumber, country_code=country_code)


def _parse_slug(val):
    return slugify(val)


def parse_slug(val):
    s = parse_str(val)
    return _parse_slug(s)


def parse_str(val):
    if type_util.is_string(val):
        try:
            val = ftfy.fix_text(val)
        except UnicodeError:
            pass
    else:
        val = text_type(val)
    val = val.strip()
    val = ' '.join(val.split())
    return val



# -*- coding: utf-8 -*-

from benedict.core.keylists import keylists
from benedict.utils import type_util


def keypaths(d, separator='.', indexes=False):
    separator = separator or '.'
    if not type_util.is_string(separator):
        raise ValueError('separator argument must be a (non-empty) string.')
    kls = keylists(d, indexes=indexes)
    kps = [separator.join(['{}'.format(key) for key in kl]) for kl in kls]
    kps.sort()
    return kps



# -*- coding: utf-8 -*-

from __future__ import absolute_import

from benedict.serializers.abstract import AbstractSerializer

import xmltodict


class XMLSerializer(AbstractSerializer):

    def __init__(self):
        super(XMLSerializer, self).__init__()

    def decode(self, s, **kwargs):
        kwargs.setdefault('dict_constructor', dict)
        data = xmltodict.parse(s, **kwargs)
        return data

    def encode(self, d, **kwargs):
        data = xmltodict.unparse(d, **kwargs)
        return data



# -*- coding: utf-8 -*-

from benedict.utils import type_util


def _merge_dict(d, other, overwrite=True):
    for key, value in other.items():
        _merge_item(d, key, value, overwrite)


def _merge_item(d, key, value, overwrite=True):
    if key in d:
        item = d.get(key, None)
        if type_util.is_dict(item) and type_util.is_dict(value):
            _merge_dict(item, value, overwrite)
        elif overwrite:
            d[key] = value
    else:
        d[key] = value


def merge(d, other, *args, **kwargs):
    overwrite = kwargs.get('overwrite', True)
    others = [other] + list(args)
    for other in others:
        _merge_dict(d, other, overwrite)
    return d



# -*- coding: utf-8 -*-

from benedict.core import clone


def filter(d, predicate):
    if not callable(predicate):
        raise ValueError('predicate argument must be a callable.')
    new_dict = clone(d, empty=True)
    keys = list(d.keys())
    for key in keys:
        value = d.get(key, None)
        if predicate(key, value):
            new_dict[key] = value
    return new_dict



# -*- coding: utf-8 -*-

from benedict.utils import type_util


def remove(d, keys, *args):
    if type_util.is_string(keys):
        keys = [keys]
    keys += args
    for key in keys:
        d.pop(key, None)



# -*- coding: utf-8 -*-

__author__ = 'Fabio Caccamo'
__copyright__ = 'Copyright (c) 2019-2020 Fabio Caccamo'
__description__ = 'python-benedict is a dict subclass with keylist/keypath support, I/O shortcuts (base64, csv, json, pickle, plist, query-string, toml, xml, yaml) and many utilities... for humans, obviously.'
__email__ = 'fabio.caccamo@gmail.com'
__license__ = 'MIT'
__title__ = 'benedict'
__version__ = '0.21.0'



# -*- coding: utf-8 -*-

from benedict.serializers import (
    get_format_by_path, get_serializer_by_format, get_serializers_extensions, )

from six import PY2

import errno
import os
import requests


def autodetect_format(s):
    if is_url(s) or is_filepath(s):
        return get_format_by_path(s)
    return None


def decode(s, format, **kwargs):
    serializer = get_serializer_by_format(format)
    if not serializer:
        raise ValueError('Invalid format: {}.'.format(format))
    decode_opts = kwargs.copy()
    data = serializer.decode(s.strip(), **decode_opts)
    return data


def encode(d, format, **kwargs):
    serializer = get_serializer_by_format(format)
    if not serializer:
        raise ValueError('Invalid format: {}.'.format(format))
    s = serializer.encode(d, **kwargs)
    return s


def is_data(s):
    return (len(s.splitlines()) > 1)


def is_filepath(s):
    if any([s.endswith(ext) for ext in get_serializers_extensions()]):
        return True
    return os.path.isfile(s)


def is_url(s):
    return any([s.startswith(protocol)
                for protocol in ['http://', 'https://']])


def read_content(s):
    # s -> filepath or url or data
    if is_data(s):
        # data
        return s
    elif is_url(s):
        # url
        return read_url(s)
    elif is_filepath(s):
        # filepath
        return read_file(s)
    # one-line data?!
    return s


def read_file(filepath, encoding='utf-8'):
    if os.path.isfile(filepath):
        content = ''
        options = {} if PY2 else { 'encoding':encoding }
        with open(filepath, 'r', **options) as file:
            content = file.read()
        return content
    return None


def read_url(url, *args, **kwargs):
    response = requests.get(url, *args, **kwargs)
    if response.status_code == requests.codes.ok:
        content = response.text
        return content
    raise ValueError(
        'Invalid url response status code: {}.'.format(
            response.status_code))


def write_file_dir(filepath):
    filedir = os.path.dirname(filepath)
    if os.path.exists(filedir):
        return
    try:
        os.makedirs(filedir)
    except OSError as e:
        # Guard against race condition
        if e.errno != errno.EEXIST:
            raise e


def write_file(filepath, content, encoding='utf-8'):
    # https://stackoverflow.com/questions/12517451/automatically-creating-directories-with-file-output
    write_file_dir(filepath)
    options = {} if PY2 else { 'encoding':encoding }
    with open(filepath, 'w+', **options) as file:
        file.write(content)
    return True



# -*- coding: utf-8 -*-

from __future__ import absolute_import

from benedict.serializers.abstract import AbstractSerializer

import yaml


class YAMLSerializer(AbstractSerializer):

    def __init__(self):
        super(YAMLSerializer, self).__init__()

    def decode(self, s, **kwargs):
        data = yaml.safe_load(s, **kwargs)
        return data

    def encode(self, d, **kwargs):
        data = yaml.dump(dict(d), **kwargs)
        return data



# -*- coding: utf-8 -*-

from benedict.core.keypaths import keypaths
from benedict.utils import type_util

import re


def match(d, pattern, separator='.', indexes=True):
    if type_util.is_regex(pattern):
        regex = pattern
    elif type_util.is_string(pattern):
        # all indexes wildcard support
        pattern = re.sub(r'([\*]{1})', '(.)*', pattern)
        # escape square brackets
        pattern = re.sub(r'(\[([^\[\]]*)\])', '\\[\\g<2>\\]', pattern)
        regex = re.compile(pattern, flags=re.DOTALL)
    else:
        raise ValueError('Expected regex or string, found: {}'.format(
            type(pattern)))
    kps = keypaths(d, separator=separator, indexes=indexes)
    values = [d.get(kp) for kp in kps if regex.match(kp)]
    return values



# -*- coding: utf-8 -*-

from benedict.serializers.abstract import AbstractSerializer
from benedict.serializers.base64 import Base64Serializer
from benedict.serializers.csv import CSVSerializer
from benedict.serializers.json import JSONSerializer
from benedict.serializers.pickle import PickleSerializer
from benedict.serializers.plist import PListSerializer
from benedict.serializers.query_string import QueryStringSerializer
from benedict.serializers.toml import TOMLSerializer
from benedict.serializers.xml import XMLSerializer
from benedict.serializers.yaml import YAMLSerializer

import re


_BASE64_SERIALIZER = Base64Serializer()
_CSV_SERIALIZER = CSVSerializer()
_JSON_SERIALIZER = JSONSerializer()
_PICKLE_SERIALIZER = PickleSerializer()
_PLIST_SERIALIZER = PListSerializer()
_QUERY_STRING_SERIALIZER = QueryStringSerializer()
_TOML_SERIALIZER = TOMLSerializer()
_YAML_SERIALIZER = YAMLSerializer()
_XML_SERIALIZER = XMLSerializer()

_SERIALIZERS = {
    'b64': _BASE64_SERIALIZER,
    'base64': _BASE64_SERIALIZER,
    'csv': _CSV_SERIALIZER,
    'json': _JSON_SERIALIZER,
    'pickle': _PICKLE_SERIALIZER,
    'plist': _PLIST_SERIALIZER,
    'qs': _QUERY_STRING_SERIALIZER,
    'querystring': _QUERY_STRING_SERIALIZER,
    'toml': _TOML_SERIALIZER,
    'yaml': _YAML_SERIALIZER,
    'yml': _YAML_SERIALIZER,
    'xml': _XML_SERIALIZER,
}

_SERIALIZERS_EXTENSIONS = [
    '.{}'.format(extension) for extension in _SERIALIZERS.keys()]


def get_format_by_path(path):
    path = path.lower()
    for extension in _SERIALIZERS_EXTENSIONS:
        if path.endswith(extension):
            return extension[1:]
    return None


def get_serializer_by_format(format):
    format_key = (format or '').lower().strip()
    format_key = re.sub(r'[\s\-\_]*', '', format_key)
    return _SERIALIZERS.get(format_key)


def get_serializers_extensions():
    return list(_SERIALIZERS_EXTENSIONS)



# -*- coding: utf-8 -*-

from benedict.core.clean import clean
from benedict.core.clone import clone
from benedict.core.dump import dump
from benedict.core.filter import filter
from benedict.core.find import find
from benedict.core.flatten import flatten
from benedict.core.groupby import groupby
from benedict.core.invert import invert
from benedict.core.items_sorted import (
    items_sorted_by_keys, items_sorted_by_values, )
from benedict.core.keylists import keylists
from benedict.core.keypaths import keypaths
from benedict.core.match import match
from benedict.core.merge import merge
from benedict.core.move import move
from benedict.core.nest import nest
from benedict.core.remove import remove
from benedict.core.rename import rename
from benedict.core.search import search
from benedict.core.standardize import standardize
from benedict.core.subset import subset
from benedict.core.swap import swap
from benedict.core.traverse import traverse
from benedict.core.unflatten import unflatten
from benedict.core.unique import unique



# -*- coding: utf-8 -*-

from benedict.core.rename import rename
from benedict.core.traverse import traverse
from benedict.utils import type_util

from slugify import slugify

import re


def _standardize_item(d, key, value):
    if type_util.is_string(key):
        # https://stackoverflow.com/a/12867228/2096218
        norm_key = re.sub(
            r'((?<=[a-z0-9])[A-Z]|(?!^)[A-Z](?=[a-z]))', r'_\1', key)
        norm_key = slugify(norm_key, separator='_')
        rename(d, key, norm_key)


def standardize(d):
    traverse(d, _standardize_item)



# -*- coding: utf-8 -*-

from benedict.utils import type_util


def _get_keylist_for_dict(d, parent_keys, indexes):
    keylist = []
    for key, value in d.items():
        keys = parent_keys + [key]
        keylist += [keys]
        keylist += _get_keylist_for_value(value, keys, indexes)
    return keylist


def _get_keylist_for_list(l, parent_keys, indexes):
    keylist = []
    for key, value in enumerate(l):
        keys = list(parent_keys)
        keys[-1] += '[{}]'.format(key)
        keylist += [keys]
        keylist += _get_keylist_for_value(value, keys, indexes)
    return keylist


def _get_keylist_for_value(value, parent_keys, indexes):
    if type_util.is_dict(value):
        return _get_keylist_for_dict(value, parent_keys, indexes)
    elif type_util.is_list(value) and indexes:
        return _get_keylist_for_list(value, parent_keys, indexes)
    return []


def keylists(d, indexes=False):
    return _get_keylist_for_value(d, [], indexes)



# -*- coding: utf-8 -*-


def _items_sorted_by_item_at_index(d, index, reverse):
    return sorted(d.items(), key=lambda item: item[index], reverse=reverse)


def items_sorted_by_keys(d, reverse=False):
    return _items_sorted_by_item_at_index(d, 0, reverse)


def items_sorted_by_values(d, reverse=False):
    return _items_sorted_by_item_at_index(d, 1, reverse)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import find_packages, setup

import os, sys

exec(open('benedict/metadata.py').read())

github_url = 'https://github.com/fabiocaccamo'
package_name = 'python-benedict'
package_url = '{}/{}'.format(github_url, package_name)
package_path = os.path.abspath(os.path.dirname(__file__))
long_description_file_path = os.path.join(package_path, 'README.md')
long_description_content_type = 'text/markdown'
long_description = ''
try:
    long_description_file_options = {} if sys.version_info[0] < 3 else { 'encoding':'utf-8' }
    with open(long_description_file_path, 'r', **long_description_file_options) as f:
        long_description = f.read()
except IOError:
    pass

setup(
    name=package_name,
    packages=find_packages(exclude=['contrib', 'docs', 'tests*']),
    include_package_data=True,
    version=__version__,
    description=__description__,
    long_description=long_description,
    long_description_content_type=long_description_content_type,
    author=__author__,
    author_email=__email__,
    url=package_url,
    download_url='{}/archive/{}.tar.gz'.format(package_url, __version__),
    keywords=[
        'python', 'dictionary', 'dictionaries', 'dict', 'benedict',
        'subclass', 'extended', 'keylist', 'keypath', 'utility', 'io',
        'data', 'file', 'url', 'read', 'write', 'parse',
        'base64', 'csv', 'json', 'query-string', 'toml', 'xml', 'yaml',
        'clean', 'clone', 'deepclone', 'deepupdate', 'dump',
        'filter', 'flatten', 'groupby', 'invert', 'merge',
        'move', 'nest', 'remove', 'rename', 'search', 'standardize',
        'subset', 'swap', 'traverse', 'unflatten', 'unique',
    ],
    install_requires=[
        'ftfy==4.4.3;python_version<"3.4"',
        'ftfy;python_version>"2.7"',
        'mailchecker',
        'phonenumbers',
        'python-dateutil',
        'python-slugify',
        'pyyaml',
        'requests',
        'six',
        'toml',
        'xmltodict',
    ],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Topic :: Software Development :: Build Tools',
    ],
    license=__license__,
    test_suite='tests'
)



# -*- coding: utf-8 -*-

from benedict.utils import type_util


def _clean_item(d, key, strings, collections):
    value = d.get(key, None)
    if not value:
        del_none = (value is None)
        del_string = (strings and type_util.is_string(value))
        del_collection = (collections and type_util.is_collection(value))
        return any([del_none, del_string, del_collection])
    return False


def clean(d, strings=True, collections=True):
    keys = list(d.keys())
    for key in keys:
        if _clean_item(d, key, strings, collections):
            del d[key]



# -*- coding: utf-8 -*-

from benedict.core import clone
from benedict.utils import type_util


def subset(d, keys, *args):
    new_dict = clone(d, empty=True)
    if type_util.is_string(keys):
        keys = [keys]
    keys += args
    for key in keys:
        new_dict[key] = d.get(key, None)
    return new_dict



# -*- coding: utf-8 -*-

from __future__ import absolute_import

from benedict.serializers.abstract import AbstractSerializer

import plistlib
import six
import unicodedata


class PListSerializer(AbstractSerializer):
    """
    https://docs.python.org/3/library/plistlib.html
    """
    def __init__(self):
        super(PListSerializer, self).__init__()

    def decode(self, s, **kwargs):
        if six.PY2:
            if isinstance(s, unicode):
                s = unicodedata.normalize('NFKD', s).encode('ascii', 'ignore')
            return plistlib.readPlistFromString(s)
        kwargs.setdefault('fmt', plistlib.FMT_XML)
        encoding = kwargs.pop('encoding', 'utf-8')
        return plistlib.loads(s.encode(encoding), **kwargs)

    def encode(self, d, **kwargs):
        if six.PY2:
            return plistlib.writePlistToString(d)
        encoding = kwargs.pop('encoding', 'utf-8')
        return plistlib.dumps(d, **kwargs).decode(encoding)



# -*- coding: utf-8 -*-

from __future__ import absolute_import

from benedict.serializers.abstract import AbstractSerializer

import toml


class TOMLSerializer(AbstractSerializer):

    def __init__(self):
        super(TOMLSerializer, self).__init__()

    def decode(self, s, **kwargs):
        data = toml.loads(s, **kwargs)
        return data

    def encode(self, d, **kwargs):
        data = toml.dumps(d, **kwargs)
        return data



# -*- coding: utf-8 -*-

from datetime import datetime
from decimal import Decimal
from six import integer_types, string_types

import re

regex = re.compile('').__class__


def is_bool(val):
    return isinstance(val, bool)


def is_collection(val):
    return isinstance(val, (dict, list, set, tuple, ))


def is_datetime(val):
    return isinstance(val, datetime)


def is_decimal(val):
    return isinstance(val, Decimal)


def is_dict(val):
    return isinstance(val, dict)


def is_dict_or_list(val):
    return isinstance(val, (dict, list, ))


def is_float(val):
    return isinstance(val, float)


def is_function(val):
    return callable(val)


def is_integer(val):
    return isinstance(val, integer_types)


def is_json_serializable(val):
    json_types = (type(None), bool, dict, float, list, tuple, ) + \
                 integer_types + string_types
    return isinstance(val, json_types)


def is_list(val):
    return isinstance(val, list)


def is_list_or_tuple(val):
    return isinstance(val, (list, tuple, ))


def is_none(val):
    return val is None


def is_not_none(val):
    return val is not None


def is_regex(val):
    return isinstance(val, regex)


def is_set(val):
    return isinstance(val, set)


def is_string(val):
    return isinstance(val, string_types)


def is_tuple(val):
    return isinstance(val, tuple)
